//-< DHASH.CPP >-----------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET   *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Dynamic extended hash table                                       *  /  \  *
//                           Created:      30-Jul-94   K.A. Knizhnik * / [] \ *
//                           Last update:  30-Jul-94   K.A. Knizhnik * GARRET *
//-------------------------------------------------------------------*--------*

#include <string.h>
#include "oodbs.hpp"
#include "dhash.hpp"

large Dbs_Str_Hash_Table::Block::Hash_Func( const char* str ) {
    tiny ch;
    large key;
    for (key = 0L; (ch=*str++) != 0; key = (key << 2) ^ (key >> 16) ^ ch);
    return key;
}

File_Object* Dbs_Str_Hash_Table::Block::Search( const char* str ) {
    char* key = Keys;
    for( int i = Numb; -- i >= 0; key += strlen(key) + 1 )
        if( strcmp( key, str ) == 0 ) return Ref(Numb - i - 1);
    return NULL;
}

bool Dbs_Str_Hash_Table::Block::Put( File_Object* obj, const char* key ) {
    int Len = strlen(key) + 1;
    if( Keys + Used + Len > (char*)&Ref(Numb) ) return FALSE;
    Modify();
    strcpy( Keys + Used, key );
    Used += Len;
    Ref(Numb++) = obj;
    return TRUE;
}

void Dbs_Str_Hash_Table::Block::Unlink( File_Object* Rec ) {
    File_Object** Recs = &Ref(-1);
    char* key;
    Modify();
    for(key = Keys; *--Recs != Rec; key += strlen(key) + 1 );
    int del = strlen(key) + 1;
    memcpy( key, key + del, Keys + (Used -= del) - key );
    for( File_Object** Last = &Ref(--Numb); Recs != Last; Recs -= 1 )
        *Recs = *(Recs-1);
}

Dbs_Str_Hash_Table::Block* Dbs_Str_Hash_Table::Block::Fork() {
    Block* Blk = new (Dbs_Size - sizeof(Block)) Block;
    large  mask = 1L << Bits;
    int    n = Numb;
    Modify();
    Clear(), Blk->Clear(), Blk->Bits = ++Bits;
    char* key = Keys;
    for( int i = 0; i < n; i ++ ) {
        int len = strlen(key) + 1;
        ((Hash_Func(key) & mask) ? this : Blk)->Put( Ref(i), key );
        key += len;
    }
    return (Block*)Blk->Dbs_Pos();
}


Dbs_Str_Hash_Table::Segment* Dbs_Str_Hash_Table::Segment::Fork() {
    Segment* Seg = new (Dbs_Size-sizeof(Segment)) Segment;
    memcpy( Seg, this, Dbs_Size );
    return (Segment*)Seg->Dbs_Pos();
}

Dbs_Str_Hash_Table::Segment_List* Dbs_Str_Hash_Table::Segment_List::Fork() {
    Dbs_Fix();
    Segment_List* New =
        new ((Dbs_Size-sizeof(File_Object))*2-sizeof(Segment*)) Segment_List;
    New->Dbs_Fix();
    Segment **Src = Ref, **Last = (Segment**)((char*)this + Dbs_Size);
    Segment **Dst = New->Ref;
    int half = Last - Src;
    while( Src < Last )
        *(Dst+half) = ((Segment*)Dbs_Load(*Src))->Fork(), *Dst++ = *Src++;
    Dbs_UnFix(), New->Dbs_UnFix();
    delete this;
    return New;
}

Dbs_Str_Hash_Table::Dbs_Str_Hash_Table( int Segment_Log_Size, int Block_Size ) {
    Seg_Log_Size = Segment_Log_Size;
    Seg_Log_Numb = 0;
    Blk_Size = ((Block_Size+Dbs_Slot_Size-1)&~(Dbs_Slot_Size-1))-sizeof(Block);
    Dbs_Fix();
    Segment_List* Lst = new Segment_List;
    Lst->Dbs_Fix();
    Segment* Seg =
        new ((sizeof(Block*) << Segment_Log_Size) - sizeof(Block*)) Segment;
    memset( Seg->Ref, 0, sizeof(Block*) << Segment_Log_Size );
    Lst->Ref[0] = (Segment*)Seg->Dbs_Pos();
    List = (Segment_List*)Lst->Dbs_Pos();
    Dbs_UnFix(), Lst->Dbs_UnFix();
}


void Dbs_Str_Hash_Table::Put( File_Object* obj, const char* str ) {
    large Key = Block::Hash_Func(str);
    Dbs_Fix();

  Rehash:
    int sn = (Key >> Seg_Log_Size) & ((1 << Seg_Log_Numb) - 1);
    int bn = Key & ((1L << Seg_Log_Size) - 1);

    Segment_List* Lst = (Segment_List*)Dbs_Load(List); Lst->Dbs_Fix();
    Segment*      Seg = (Segment*)Dbs_Load(Lst->Ref[sn]);
    Block*        Blk = Seg->Ref[bn];

    if( Blk == NULL ) {
        Blk = new ((o_size_t)Blk_Size) Block;
        Blk->Clear(), Blk->Bits = Seg_Log_Numb + Seg_Log_Size;
        Seg->Modify();
        Seg->Ref[bn] = (Block*)Blk->Dbs_Pos();
    } else Blk = (Block*)Dbs_Load(Blk);
    Blk->Dbs_Fix();

    if( !Blk->Put( obj, str ) ) {
        int ds;
        if( Blk->Bits == Seg_Log_Numb + Seg_Log_Size ) {
            ds = 1 << Seg_Log_Numb;
            Lst->Dbs_UnFix(), Lst = Lst->Fork(), Lst->Dbs_Fix();
            Modify(), Seg_Log_Numb += 1, List = (Segment_List*)Lst->Dbs_Pos();
        } else ds = 1 << (Blk->Bits - Seg_Log_Size), sn &= ds - 1;
        Block* Forked = Blk->Fork();
        int n = 1 << Seg_Log_Numb;
        ds <<= 1;
        do {
            Seg = (Segment*)Dbs_Load(Lst->Ref[sn]);
            Seg->Modify(), Seg->Ref[bn] = Forked;
        } while( (sn += ds) < n );
        Lst->Dbs_UnFix(), Blk->Dbs_UnFix();
        goto Rehash;
    }
    Dbs_UnFix(), Lst->Dbs_UnFix(), Blk->Dbs_UnFix();
}

File_Object* Dbs_Str_Hash_Table::Search( const char* str ) {
    large Key = Block::Hash_Func(str);

    int sn = (Key >> Seg_Log_Size) & ((1 << Seg_Log_Numb) - 1);
    int bn = Key & ((1L << Seg_Log_Size) - 1);

    Segment_List* Lst = (Segment_List*)Dbs_Load(List);
    Segment*      Seg = (Segment*)Dbs_Load(Lst->Ref[sn]);
    Block*        Blk = Seg->Ref[bn];

    return Blk ? ((Block*)Dbs_Load(Blk))->Search(str) : (File_Object*)0;
}

void Dbs_Str_Hash_Table::Unlink( File_Object* obj, const char* str ) {
    large Key = Block::Hash_Func(str);

    int sn = (Key >> Seg_Log_Size) & ((1 << Seg_Log_Numb) - 1);
    int bn = Key & ((1L << Seg_Log_Size) - 1);

    Segment_List* Lst = (Segment_List*)Dbs_Load(List);
    Segment*      Seg = (Segment*)Dbs_Load(Lst->Ref[sn]);
    Block*        Blk = (Block*)Dbs_Load(Seg->Ref[bn]);

    Blk->Unlink(obj);
}
