//-< FSTOR.CPP >-----------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET   *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Store mapped on file                                              *  /  \  *
//                           Created:      04-Mar-92   K.A. Knizhnik * / [] \ *
//                           Last update:  16-May-92   K.A. Knizhnik * GARRET *
//-------------------------------------------------------------------*--------*

#include <stdio.h>
#include <string.h>

#include "oodbs.hpp"
#include "fstor.hpp"

import tiny Dbs_Max_Hole_Offset [];
import tiny Dbs_First_Hole_Size [];
import tiny Dbs_Last_Hole_Size  [];
import tiny Dbs_Max_Hole_Size   [];

Dbs_F_Store::Dbs_F_Store(Dbs_File&St_File,Dbs_Pool&St_Pool):Pool_Store(St_Pool){
  File = &St_File;
  Page_Size = Pool->Quant ();
  Page_Self = 1 << ( Pool->Power() - Dbs_Slot_Bits - 3 );
}

void Dbs_F_Store::Open ( int Id, Object_Database* Db ) {
  Ident = Id, Base = Db;
  Dbs_Check ( File->Open( O_RDWR|O_CREAT|O_BINARY, S_IREAD|S_IWRITE ) != -1 );
  large Length = File->Length();
  if ( Length == 0 )  {   // prepare virtual allocation mechanism
     Dbs_Page* Pg = Pool->New ( this, Last_Page = 0 );
     memset ( Pg->Space, 0, Page_Size );
     *(nat*)Pg->Space = Page_Self;
  } else {
     Last_Page = Length & ~( (large(Page_Size) << (3 + Dbs_Slot_Bits)) - 1 );
  }
  Bit_Base = Page_Self;
  Bit_Page = 0;
}

large Dbs_F_Store::Length( void )  { return File->Length(); }
large Dbs_F_Store::Start ( void )  { return Page_Size; }
bool  Dbs_F_Store::Empty ( void )  { return Length() <= Start(); }
void  Dbs_F_Store::Flush ( void )  { Pool->Flush ( this ); }

void  Dbs_F_Store::Close ( void )
  { Pool->Flush(this); Dbs_Check( File->Close() == Ok ); }

void  Dbs_F_Store::Save  ( Dbs_Page* Pg )  {
  File->Seek ( Pg->Address , 0 );
  Dbs_Check  ( (unsigned)File->Write ( Pg->Space , Page_Size ) == Page_Size );
}

void  Dbs_F_Store::Load  ( Dbs_Page* Pg )  {
  if ( Pg->Address < File->Length() )  {
     File->Seek ( Pg->Address , 0 );
     Dbs_Check  ( (unsigned)File->Read( Pg->Space, Page_Size ) == Page_Size );
  }
}

//
// Read / write objects in pool
//

void Dbs_F_Store::Read ( Dbs_Object* Obj ) {
  unsigned Page_Size = this->Page_Size;
  large    Addr = Odb_Offset( Obj->Address );
  unsigned Offs = unsigned(Addr) & (Page_Size-1);
  large    Page = Addr - Offs;

  Dbs_Page* Pg = Pool->Find ( this, Page );
  o_size_t  Rest, Size = ((File_Object*)(Pg->Space + Offs))->Dbs_Size;

  if( Size > Obj->Size ) Base->Realloc( Obj, Size );
  Obj->Store = this;

  char* Buffer = Obj->Buffer;

  while ( (Rest = Page_Size - Offs) < Size )  {
     memcpy ( Buffer, Pg->Space + Offs, Rest );
     Pg = Pool->Find ( this, Page += Page_Size );
     Buffer += Rest; Size -= Rest;
     Offs = 0 ;
  }
  memcpy ( Buffer, Pg->Space + Offs, Size );
}

void Dbs_F_Store::Write ( Dbs_Object* Obj ) {
  tiny*    Buffer = (tiny*)Obj->Buffer;
  large    Addr = Odb_Offset( Obj->Address );
  o_size_t Size = ((File_Object*)Buffer)->Dbs_Size;
  unsigned Page_Size = this->Page_Size;
  unsigned Offs = unsigned(Addr) & (Page_Size-1);
  large    Page = Addr - Offs;

  while ( TRUE )  {
     Dbs_Page* Pg = Pool->Find ( this, Page );
     Pg->State = Pg_Dirty;
     o_size_t Rest = Page_Size - Offs;
     if ( Size <= Rest )  {
        memcpy ( Pg->Space + Offs, Buffer, Size );
        return;
     }
     memcpy ( Pg->Space + Offs, Buffer, Rest );
     Buffer += Rest, Size -= Rest;
     Page += Page_Size;
     Offs = 0;
  }
}

//
// Memory allocation mechanism
//

void  Dbs_F_Store::Near ( large Address )  {
  Bit_Page = Address & -long(large(Page_Size) << (Dbs_Slot_Bits + 3));
  Bit_Base = unsigned ( ( Address - Bit_Page ) >> ( Dbs_Slot_Bits + 3 ) );
  while ( Bit_Page > Last_Page ) {
     Last_Page += large(Page_Size) << (3 + Dbs_Slot_Bits);
     Dbs_Page* Pg = Pool->New ( this, Last_Page );
     memset( Pg->Space, 0, Page_Size );
     *(nat*)Pg->Space = Page_Self;
  }
}

local void Reserve ( Dbs_Page* Pg, large Bit_Ofs, unsigned Bit_Size )  {
  tiny* Ptr = Pg->Space + unsigned( Bit_Ofs >> 3 );
  unsigned Offs = unsigned(Bit_Ofs) & 7;
  Pg->State = Pg_Dirty;
  if ( Bit_Size + Offs > 8 )  {
     *Ptr++ |= -1 << Offs;
     Bit_Size -= 8 - Offs;
     for (; Bit_Size >= 8; Bit_Size -= 8 ) *Ptr++ = ~0;
     *Ptr |= (1 << Bit_Size) - 1;
  }
  else *Ptr |= ((1 << Bit_Size)-1) << Offs;
}

large Dbs_F_Store::Malloc( o_size_t Size )  {
  Dbs_Page* Pg;
  unsigned  Space, Bit_Base = this->Bit_Base;
  unsigned  Bit_Size = ( Size + Dbs_Slot_Size - 1 ) >> Dbs_Slot_Bits;
  unsigned  Page_Size = this->Page_Size;
  large     Slot = large(Bit_Base) << 3;
  large     Segm = large(Page_Size) << (3 + Dbs_Slot_Bits);
  tiny*     Point;

  Pg = Pool->Find ( this, Bit_Page );
  Point = Pg->Space + Bit_Base;
  for ( Space=0; Bit_Base < Page_Size; Point++ , Bit_Base++ ) {
        if ( Space + Dbs_First_Hole_Size [ *Point ] >= Bit_Size )  {
        Reserve ( Pg , Slot , Bit_Size ); this->Bit_Base = Bit_Base;
        return  Bit_Page + (Slot << Dbs_Slot_Bits);
     } else if ( unsigned(Dbs_Max_Hole_Size[ *Point ]) >= Bit_Size )  {
        Slot += Space + Dbs_Max_Hole_Offset [ * Point ];
        Reserve ( Pg , Slot , Bit_Size ); this->Bit_Base = Bit_Base;
        return  Bit_Page + (Slot << Dbs_Slot_Bits);
     } else if ( Dbs_Last_Hole_Size [ *Point ] < 8 )  {
        Slot += Space;
        Space = Dbs_Last_Hole_Size [ *Point ];
        Slot += 8 - Space;
     } else Space += 8;
  }
  Bit_Page = 0;
  do {
     Pg = Pool->Find ( this, Bit_Page );
     Bit_Base = *(nat*)Pg->Space;
     Point = Pg->Space + Bit_Base;
     while ( Bit_Base < Page_Size && *(Point++) == 0xFF)  Bit_Base += 1;
     *(nat*)Pg->Space = Bit_Base;
     Point -= 1;
     Slot = large(Bit_Base) << 3;
     for ( Space = 0; Bit_Base < Page_Size; Point++, Bit_Base++ ) {
        if ( Space + Dbs_First_Hole_Size [ *Point ] >= Bit_Size ) {
           Reserve ( Pg, Slot, Bit_Size ); this->Bit_Base = Bit_Base;
           return  Bit_Page + (Slot << Dbs_Slot_Bits);
        } else if ( unsigned(Dbs_Max_Hole_Size[*Point]) >= Bit_Size ) {
           Slot += Space + Dbs_Max_Hole_Offset [ * Point ];
           Reserve ( Pg, Slot, Bit_Size ); this->Bit_Base = Bit_Base;
           return  Bit_Page + ( Slot << Dbs_Slot_Bits );
        } else if ( Dbs_Last_Hole_Size [ *Point ] < 8 ) {
           Slot += Space;
           Space = Dbs_Last_Hole_Size [ *Point ];
           Slot += 8 - Space;
        } else Space += 8;
     }
  } while ( ( Bit_Page += Segm ) <= Last_Page );
  Pg = Pool->New ( this, Last_Page = Bit_Page );
  Point = Pg->Space;
  memset ( Point , 0 , Page_Size );
  Point += Bit_Base = Page_Self;
  for (; Bit_Size >= 8; Bit_Size -= 8 , Bit_Base ++ )  *(Point++) = ~0;
  *Point = (1 << Bit_Size) - 1;
  *(nat*)Pg->Space = this->Bit_Base = Bit_Base;
  return  Bit_Page + Page_Size ;
}

void Dbs_F_Store::Free ( large Address, o_size_t Size )  {
  unsigned Bit_Size = ( Size + Dbs_Slot_Size - 1 ) >> Dbs_Slot_Bits;
  large    Page = Address & -long(large(Page_Size) << (Dbs_Slot_Bits + 3));
  large    Bit  = ( Address - Page ) >> Dbs_Slot_Bits;
  tiny     *Point, Mask = 1 << ( Bit & 7 );
  unsigned Base = unsigned(Bit >> 3);

  Dbs_Page *Pg = Pool->Find ( this, Page );
  Pg->State = Pg_Dirty;

  if ( unsigned(*(nat*)Pg->Space) > Base )  *(nat*)Pg->Space = Base;
  for ( Point = Pg->Space + Base; Bit_Size > 0; Bit_Size -- )  {
     * Point &= ~Mask;
     if ( ( Mask <<= 1 ) == 0 )  {
        Mask = 1;
        Point ++;
     }
  }
}


void Dbs_F_Store::Clear () {
  large Length = File->Length();
  large Segm = large(Page_Size) << (Dbs_Slot_Bits + 3);
  for ( large Page = 0; Page < Length; Page += Segm ) {
     Dbs_Page* Pg = Pool->Find ( this, Page );
     memset ( Pg->Space, 0, Page_Size );
     *(nat*)Pg->Space = Page_Self;
     Pg->State = Pg_Dirty;
  }
  Bit_Base = Page_Self;
}
