//-------------------------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET   *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Low level multi file I/O operations                               *  /  \  *
//                           Created:      15-Feb-92   K.A. Knizhnik * / [] \ *
//                           Last update:  24-Oct-92   K.A. Knizhnik * GARRET *
//-------------------------------------------------------------------*--------*

#include <stdio.h>
#include "mflio.hpp"

Dbs_M_File::Dbs_M_File( nat Number , Dbs_Segment * Segment )  {
  ASSERT ( Number > 0 );
  Seg_Num = Number;
  Limit = new large [ Number ];
  Seg   = new Dbs_File* [ Number ];
  for ( int i = Number; i --; )   {
     Limit[i] = Segment[i].Limit;
     Seg[i] = new Dbs_File ( Segment[i].Name );
  }
  Name = NULL;
}

Dbs_M_File::~Dbs_M_File ( void ) {
  for ( int i = Seg_Num; i --; delete Seg[i] );
  delete[] Limit; delete[] Seg;
}

int Dbs_M_File::Open ( int Flags , int Mode )  {
  large File_Size = 0;
  for ( int i = Seg_Num; i --; )  {
     if ( Seg[i]->Open ( Flags , Mode ) == -1 ) return -1;
     large Seg_Size = Seg[i]->Last_Pos;
     if ( Seg_Size > Limit[i] || (File_Size > 0 && Seg_Size != Limit[i]) )
        Abzac ( "Illegal length of segment '%s' !", Seg[i]->Name );
     File_Size += Seg_Size;
  }
  Last_Pos = File_Size;
  Curr_Pos = Curr_Ofs = 0;
  Curr_Ind = 0;
  return Ok;
}

int  Dbs_M_File::Close ( void )  {
  for ( int i = Seg_Num; i --; Seg[i]->Close () );
  return Ok;
}


int  Dbs_M_File::Access ( int Mode )  {
  for ( int i=Seg_Num; i--; ) if ( Seg[i]->Access(Mode) != Ok ) return -1;
  return Ok;
}

int Dbs_M_File::Read ( void* Buffer, unsigned Quant )  {
  if ( Curr_Pos + Quant > Last_Pos )  Quant = unsigned(Last_Pos - Curr_Pos);
  unsigned Size = Quant;
  while ( Curr_Ofs + Size > Limit[Curr_Ind] )  {
     unsigned Portion = unsigned(Limit[Curr_Ind] - Curr_Ofs);
     if ( (unsigned)Seg[Curr_Ind]->Read ( Buffer , Portion ) != Portion )  { 
        Io_Error();
     }
     Buffer = (char*)Buffer + Portion;
     Size -= Portion;
     Curr_Ofs = 0;
     Curr_Pos += Portion;
     Curr_Ind ++;
     Seg[Curr_Ind]->Seek ( 0 , 0 );
  }
  if ((unsigned) Seg[Curr_Ind]->Read ( Buffer , Size ) != Size )  Io_Error();
  Curr_Ofs += Size;
  Curr_Pos += Size;
  return Quant;
}

int Dbs_M_File::Write ( void * Buffer , unsigned Quant )  {
  unsigned Size = Quant;
  if ( Curr_Pos + Size > Last_Pos )  Last_Pos = Curr_Pos + Size;
  while ( Curr_Ofs + Size > Limit[Curr_Ind] )  {
     unsigned Portion = unsigned(Limit[Curr_Ind] - Curr_Ofs);
     if ( (unsigned)Seg[Curr_Ind]->Write ( Buffer , Portion ) != Portion ) { 
        Io_Error();
     }
     Buffer = (char*)Buffer + Portion;
     Size -= Portion;
     Curr_Pos += Portion;
     if ( ++Curr_Ind < Seg_Num )  {
        Curr_Ofs = 0;
        Seg[Curr_Ind]->Seek ( 0 , 0 );
     } else {
        Curr_Ind --;
        Last_Pos -= Size;
        return Quant - Size;
     }
  }
  if ( (unsigned)Seg[Curr_Ind]->Write ( Buffer , Size ) != Size )  Io_Error();
  Curr_Ofs += Size;
  Curr_Pos += Size;
  return Quant;
}

large Dbs_M_File::Seek ( long Offset , int Whence )  {
  int i;
  if      ( Whence == 1 )  Offset += Curr_Pos;
  else if ( Whence == 2 )  Offset  = Last_Pos - Offset;
  Curr_Pos = Offset;
  for ( i = 0; large(Offset) > Limit[i]; i ++ )  Offset -= Limit[i];
  Curr_Ind = i;
  Curr_Ofs = Offset;
  Seg[i]->Seek ( Offset , 0 );
  return Curr_Pos;
}

int   Dbs_M_File::Unlink ( void ) {
  for ( int i = Seg_Num; i --; ) if ( Seg[i]->Unlink() != Ok ) return -1;
  return Ok;
}

#ifdef __SYSV__
  void Dbs_M_File::Sync () { Dbs_D_File::Sync(); }
#else
  void Dbs_M_File::Sync () { for ( int i = Seg_Num; i --; Seg[i]->Sync() ); }
#endif
