//-< OODBS.HPP >-----------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET   *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Virtual objects level                                             *  /  \  *
//                           Created:      14-Feb-92   K.A. Knizhnik * / [] \ *
//                           Last update:  16-May-92   K.A. Knizhnik * GARRET *
//-------------------------------------------------------------------*--------*

#ifndef  __OODBS_HPP__
 #define  __OODBS_HPP__

 #include "store.hpp"
 #include "l2lst.hpp"
 #include "hasht.hpp"
 #include "tsync.hpp"

 #define MINIDB_VERSION 1.21

 #if DEBUG_LEVEL >= DEBUG_CHECK 

 #define  Odb_Check_Bit          (1L << (sizeof(long)*8-1))
 #define  Odb_Sizeof(OBJ)        ((sizeof(OBJ)+Dbs_Slot_Size-1)&-Dbs_Slot_Size)
 #define  Odb_File(ADDR)         (int(ADDR) & (Dbs_Slot_Size-1))
 #define  Odb_Address(FILE,OFFS) (Odb_Check_Bit | (FILE) | (OFFS))
 #define  Odb_Offset(ADDR)       (large(ADDR)&~Odb_Check_Bit&-Dbs_Slot_Size)

 #else
 
 #define  Odb_Sizeof(OBJ)        ((sizeof(OBJ)+Dbs_Slot_Size-1)&-Dbs_Slot_Size)
 #define  Odb_File(ADDR)         (int(ADDR) & (Dbs_Slot_Size-1))
 #define  Odb_Address(FILE,OFFS) ((FILE) | (OFFS))
 #define  Odb_Offset(ADDR)       (large(ADDR) & -Dbs_Slot_Size)

 #endif 

 #define  Odb_Tag(CLASS)         ___##CLASS##_Tag
 #define  Odb_Initializer(CLASS) ___##CLASS##_Initializer

 #define  Odb_Get( Ptr )         (*((Dbs_Object**)(Ptr)-1))

 struct   File_Object;

 typedef  File_Object*  Virtual;
 
 #define  Physical      void*

 typedef  void (*V_Table) ( Physical );

#if defined(USE_VIRTUAL_FUNCTIONS) && !defined(TYPE_INFO)
#define TYPE_INFO 1
#endif

 #define    Odb_File_Object( CLASS )                              \
   CLASS& operator ()  ( CLASS* Vadr )                            \
     { return *(CLASS*)Find( large(Vadr), Dbs_Object::Read ); }   \
   CLASS& operator []  ( CLASS* Vadr )                            \
     { return *(CLASS*)Find( large(Vadr), Dbs_Object::Write ); }  \
   friend CLASS* Fix   ( CLASS* Ptr )                             \
     { CS(this); Odb_Get(Ptr)->Fixed += 1; return Ptr; }          \
   friend CLASS* UnFix ( CLASS* Ptr )                             \
     { CS(this); Odb_Get(Ptr)->Fixed -= 1; return Ptr; }          \
   friend CLASS* Pos   ( CLASS* Ptr )                             \
     { return (CLASS*) Odb_Get(Ptr)->Address; }

 #ifdef TYPE_INFO
     #define Odb_Constructor( CLASS ) CLASS() { Dbs_Ident = Odb_Tag(CLASS); }
 #else
     #define Odb_Constructor( CLASS )
 #endif

 #define  Odb_Define_Initializer( CLASS ) \
    local void Odb_Initializer(CLASS) (Physical Addr) { (void)new(Addr) CLASS; }

 //
 //   Interaction with stores
 //

 class Dbs_File;

 class File_Manager {
   protected :
    Dbs_Store*      Current_Store;
    int             Stores_Number;
    Dbs_Store**     Store;

   public :
   //
   // Current store store reqestes
   //
    Virtual  Start  ( void ) {
       return Virtual( Odb_Address(
          Current_Store->Ident, Current_Store->Start() ) );
    }
    large    Length ( void )   { return Current_Store->Length(); }
    bool     Empty  ( void )   { return Current_Store->Empty (); }
    void     With   ( Dbs_Store& Stor) { Current_Store = & Stor; }
    void     With   ( int Number) { Current_Store=Store[Number]; }

    File_Manager    ( int Number , Dbs_Store ** Stores ) {
      Store         = Stores  ;
      Stores_Number = Number  ;
      Current_Store = Stores[0];
    }
 };

 //
 //   Object cache mechanism
 //

 class  Object_Cache;

 class Dbs_Object : public L2_Elem {
   public:
    Hash_Elem( Dbs_Object );
    large       Address;
    char*       Buffer;
    Dbs_Store*  Store;
    o_size_t    Size;

    tiny        Fixed;
    tiny        State;

    enum { Read, Write }; // state of object

    void        Alloc ( o_size_t Size );
    void        Remove();
    void        Save  (); 
    void        Modify();
    void        Fix();
    void        UnFix();
 };

 class Object_Cache : public File_Manager {
     friend class Dbs_Object;
   protected :
    Dbs_Object  Header;        // chain of used cache elements
    Dbs_Object* Free_Chain;    // chain of free elements in cache
    V_Table*    Init_Table;    // initialize table
    Dbs_Object* Cache;         // cache array
    large       Total;         // total size of cache elements
    mutex_t     mutex;

    Dbs_Object* Hash_Table[ 256 ];

    void        Remove ( Dbs_Object* Obj );

   public:
    Physical    Find   ( large Address, tiny Mode = Dbs_Object::Read );
    void        Realloc( Dbs_Object* Obj, o_size_t Size );

    void        Flush( void ); // save changed objects in stores
    void        Clear( void ); // save objects and clear cache
    void        Throw( Physical Address ) { Remove( Odb_Get(Address) ); }
                // throw out object from cache
    void        Compaction();  // throw out as much cache objects as possible

    void        Free  ( Physical Address );
    void        Near  ( Virtual  Address );
    Physical    Malloc( Dbs_Store* Stor, o_size_t Size );
    Physical    Malloc( o_size_t Size ) { return Malloc ( Current_Store, Size ); }

    Object_Cache ( int Number, Dbs_Store** Stores, V_Table* Table );
    ~Object_Cache( void ) { delete[] Cache; }
 };

 //
 //   Object database common class
 //

 class Object_Database : public Object_Cache {
  public:
   static  Object_Database* Default;

   void    Open   ( void );
   void    Close  ( void );

   Object_Database( int Number, Dbs_Store** Stores, V_Table* Table = NULL )
       : Object_Cache ( Number, Stores, Table ) { Default = this; }
 };

  inline void Dbs_Object::Save () { 
      if( State & Write ) { 
	  CS(Store->Base);
	  State &= ~Write; 
	  Store->Write( this ); 
      }
  }

  inline void Dbs_Object::Modify()  {
      CS(Store->Base);
      State |= Dbs_Object::Write; 
  }
   
  inline void Dbs_Object::Fix() { 
      CS(Store->Base);
      Fixed += 1;
  } 
   
  inline void Dbs_Object::UnFix() { 
      CS(Store->Base);
      Fixed -= 1;
  } 

 //
 //   Base  class for file objects
 //

#ifdef USE_VIRTUAL_FUNCTIONS
 struct File_Object_Base {
     virtual void dummy_virtual_function() {}
 };
 struct File_Object : File_Object_Base {
#else
 struct File_Object {
#endif
   o_size_t Dbs_Size;  // size of object in bytes

   #ifdef TYPE_INFO
     short  Dbs_Ident; // unical identifier of object class
     File_Object()   { Dbs_Ident = -1; }
   #endif

   void Modify() { Odb_Get(this)->Modify(); }
   void Save()   { Odb_Get(this)->Save(); }

   Physical operator new ( size_t Size )
     { return Object_Database::Default->Malloc ( Size ); }
   void   operator delete ( Physical Address )
     { Odb_Get(Address)->Store->Base->Free( Address ); }
   Physical operator new ( size_t, Physical Address )
     { return Address; }
   Physical operator new ( size_t Size, o_size_t Space )
     { return Object_Database::Default->Malloc ( Size + Space ); }
   Physical operator new ( size_t Size, Dbs_Store &Store )
     { return Object_Database::Default->Malloc ( &Store, Size ); }
   Physical operator new ( size_t Size, Dbs_Store &Store, o_size_t Space )
     { return Object_Database::Default->Malloc ( &Store, Size + Space ); }

   void          Dbs_Fix() { Odb_Get(this)->Fix(); }
   void          Dbs_UnFix() { Odb_Get(this)->UnFix(); }
   Virtual       Dbs_Pos() { return Virtual(Odb_Get(this)->Address); }

  protected:
   Physical Dbs_Load( Virtual Addr )
     { return Odb_Get(this)->Store->Base->Find( large(Addr) ); }
 };

 #ifdef SMART_POINTERS
   extern Dbs_Object* const Odb_Nil_Ptr;

   #ifdef NO_TEMPLATES

     #define Dbs_Pointers_Declare(T)                                          \
     class T;                                                                 \
     class T##_Tie { /* RAM pointer */                                        \
       friend class T##_Ptr;                                                  \
       private:                                                               \
         T*   Addr;                                                           \
         void Reset()     { Addr = (T*)(&Odb_Nil_Ptr + 1); }                  \
         void Fix()       { Odb_Get(Addr)->Fix(); }                           \
         void UnFix()     { Odb_Get(Addr)->UnFix; }                           \
       public:                                                                \
         T*   operator->(){ return  Addr; }                                   \
         T&   operator* (){ return *Addr; }                                   \
                                                                              \
         friend bool Nil( T##_Tie const& P )                                  \
           { return  P.Addr == (T*)(&Odb_Nil_Ptr + 1); }                      \
         void        Set_Nil() { UnFix(); Reset(); }                          \
         inline friend void Delete( T##_Tie& );                               \
                                                                              \
         bool operator == ( T##_Tie const& P ) { return Addr == P.Addr; }     \
         bool operator != ( T##_Tie const& P ) { return Addr != P.Addr; }     \
         void operator =  ( T##_Tie const& P ) { UnFix(),Addr=P.Addr,Fix(); } \
                                                                              \
         T##_Tie() { Reset(); }                                               \
         T##_Tie( T##_Ptr const& P );                                         \
         T##_Tie( T##_Tie const& P ) { Addr = P.Addr, Fix(); }                \
         T##_Tie( T const* A ) { Addr = (T*)A; Fix(); }                       \
         ~T##_Tie() { UnFix(); }                                              \
     };                                                                       \
                                                                              \
     class T##_Ptr {                                                          \
       friend class T##_Tie;                                                  \
       private:                                                               \
         T*     Addr;                                                         \
       public:                                                                \
         T##_Tie operator->();                                                \
                                                                              \
         friend bool Nil( T##_Ptr const& P ) { return P.Addr == NULL; }       \
         void        Set_Nil()   { Addr = NULL; }                             \
                                                                              \
         bool operator == ( T##_Ptr const& P ) { return Addr == P.Addr; }     \
         bool operator != ( T##_Ptr const& P ) { return Addr != P.Addr; }     \
                                                                              \
         operator T*() { return Addr; }                                       \
         T##_Ptr( T const* A )                                                \
           { Addr = A ? (T*) Odb_Get(A)->Address : (T*) NULL; }               \
         T##_Ptr( T##_Tie const& P ) { Addr = (T*)Odb_Get(P.Addr)->Address; } \
         T##_Ptr( T##_Ptr const& P ) { Addr = P.Addr; }                       \
         T##_Ptr() {}                                                         \
                                                                              \
         static T##_Ptr Root() {                                              \
             T##_Ptr P;                                                       \
             P.Addr = (T*)Object_Database::Default->Start();                  \
             return P;                                                        \
         }                                                                    \
     };                                                                       \
     inline T##_Tie::T##_Tie( T##_Ptr const& P )                              \
       { Addr = (T*) Object_Database::Default->Find(large(P.Addr)); Fix(); }  \
     inline T##_Tie T##_Ptr::operator->() { return *this; }

     #define Dbs_Pointers_Define(T)                                           \
       inline void Delete( T##_Tie& P )                                       \
         { P.UnFix(); delete P.Addr; P.Reset(); }

  #else

     template<class T>
     class Tie { // RAM pointer
       private:
         T*   Addr;
         void Reset()     { Addr = (T*)(&Odb_Nil_Ptr + 1); }
         void Fix()       { Odb_Get(Addr)->Fix(); }
         void UnFix()     { Odb_Get(Addr)->UnFix(); }
       public:
         T*   operator->() const { return  Addr; }
         T&   operator* () const { return *Addr; }

         void        Set_Nil()       { UnFix(); Reset(); }
         friend bool Nil( Tie const& P )
           { return P.Addr == (T*)(&Odb_Nil_Ptr + 1); }
         friend void Delete( Tie& P ) {P.UnFix(); delete P.Addr; P.Reset();}

         bool operator == ( Tie const& P ) const { return Addr == P.Addr; }
         bool operator != ( Tie const& P ) const { return Addr != P.Addr; }
         void operator =  ( Tie const& P ) { UnFix(), Addr=P.Addr, Fix(); }

         Tie() { Reset(); }
         Tie( Tie const& P ) { Addr = P.Addr, Fix(); }
         Tie( T const* A ) { Addr = (T*) A; Fix(); }
         ~Tie() { UnFix(); }
     };

     template<class T>
     class Ptr {
       private:
         T*     Addr;
       public:
         Tie<T> operator->() const {
	     return Tie<T>((T*)Object_Database::Default->Find(large(Addr)));
	 }
	 operator Tie<T>() const { 
	     return Tie<T>((T*)Object_Database::Default->Find(large(Addr))); 
	 }

         friend bool   Nil( Ptr const& P ) { return P.Addr == NULL; }
         void          Set_Nil()     { Addr = NULL; }

         bool operator == ( Ptr const& P ) const { return Addr == P.Addr; }
         bool operator != ( Ptr const& P ) const { return Addr != P.Addr; }

         operator T*() const { return Addr; }
         Ptr( T const* A ) { Addr = A ? (T*)Odb_Get(A)->Address : (T*)0; }
         Ptr( Tie<T> const& P ) { Addr = (T*)Odb_Get(P.operator->())->Address;}
         Ptr( Ptr const& P ) { Addr = P.Addr; } 
         Ptr() {}

         static Ptr Root()
           { Ptr P; P.Addr=(T*)Object_Database::Default->Start(); return P; }
     };

   #endif
 #endif

#endif


