//-< TSYNC.HPP >-----------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET   *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Synchronization primitives                                        *  /  \  *
//                           Created:      26-Nov-99   K.A. Knizhnik * / [] \ *
//                           Last update:  26-Nov-99   K.A. Knizhnik * GARRET *
//-------------------------------------------------------------------*--------*

#ifndef __TSYNC_H__
#define __TSYNC_H__

#ifdef MULTITHREADED_SUPPORT
#if defined(_WIN32)
    #include <windows.h>
    class mutex_t { 
      private:
	CRITICAL_SECTION cs;
      public:
	void enter() { EnterCriticalSection(&cs); }
	void leave() { EnterCriticalSection(&cs); }

	mutex_t()    { InitializeCriticalSection(&cs); }
	~mutex_t()   { DeleteCriticalSection(&cs); } 
    };
    class critical_section { 
      private:
	mutex_t& mutex;
      public:
	critical_section(mutex_t& cs) : mutex(cs) { 
	    cs.enter(); 
	}
	~critical_section() { mutex.leave(); }
    };
    #define CS(db) critical_section __cs((db)->mutex)
#else
    #include <pthread.h>
    class mutex_t { 	
      private:
	pthread_mutex_t cs;
	int             count;
	pthread_t       owner;
      public:
	void enter() { 
	    pthread_t self = pthread_self();
	    if (owner != self) { 
		pthread_mutex_lock(&cs); 
		owner = self;
	    }
	    count += 1;
	}
	void leave() {
	    if (--count == 0) {
		owner = 0;
		pthread_mutex_unlock(&cs);
	    } 
	}
	mutex_t()  { pthread_mutex_init(&cs, NULL); }
	~mutex_t() { pthread_mutex_destroy(&cs); }
    };
    
    class critical_section { 
      private:
	mutex_t& mutex;
      public:
	critical_section(mutex_t& cs) : mutex(cs) { 
	    cs.enter();
	}
	~critical_section() { mutex.leave(); }
    };
    #define CS(db) critical_section __cs((db)->mutex)
#endif
#else
    class mutex_t {
      public:
	void enter();
	void leave();
    };
    #define CS(db) 
#endif

#endif
